#
# Partly from:
# $FreeBSD: src/sys/boot/i386/pmbr/pmbr.s,v 1.2 2007/11/26 21:29:59 jhb Exp $
#-
# Copyright (c) 2007 Yahoo!, Inc.
# All rights reserved.
# Written by: John Baldwin <jhb@FreeBSD.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the author nor the names of any co-contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#

#
# Partly from:
# $FreeBSD: src/sys/boot/i386/mbr/mbr.s,v 1.7 2004/08/28 08:39:35 yar Exp $
#
# Copyright (c) 1999 Robert Nordier
# All rights reserved.
#
# Redistribution and use in source and binary forms are freely
# permitted provided that the above copyright notice and this
# paragraph and the following disclaimer are duplicated in all
# such forms.
#
# This software is provided "AS IS" and without any express or
# implied warranties, including, without limitation, the implied
# warranties of merchantability and fitness for a particular
# purpose.
#

#
# "Hybridisation" and modifications for booting Haiku by Andre' Braga
# (andrebraga@gmail.com) The modifications herein contained are released into
# the Public Domain.
#

#
# A 512 byte PMBR boot loader that looks for the UUID of a Haiku GPT partition
# and boots it, falling back to MBR partitions if said UUID isn't found or if
# the (primary) GPT is corrupted or non-existent.
# 		Its usefulness is in being a generalised IPL that supports both
# partitioning styles, allowing it to be used transparently and even allowing
# disks to be converted from one partitioning style to the other with ease,
# should the need for more partitions or volumes over 2TiB arise (for instance
# when cloning an older disk to a newer, more capacious one).
# 		It also paves the way for Haiku to create and support booting from
# multiple volumes larger than 2TiB, which we're in the very privileged position
# of enjoying efficiently in the near future due to BFS. Another use case is
# taking a disk from a Intel EFI machine, plugging it on a BIOS machine and boot
# just fine; and vice-versa.
# 		As mentioned, if there are valid partitions defined in the MBR, and the
# primary GPT becomes corrupt, it can fall back to loading the MBR partition
# with the active flag set, if one is defined.
# 		Currently there's no provision for falling back to the GPT copy that
# lives in the end of the disk, due to the 512 bytes constraint; supporting this
# would require either some major refactoring of the code or disabling the
# support for booting from MBR using calls to int13h other than function 4200h,
# "LBA Read from Disk".
# 		For the GPT boot, the number of sectors to load from the partition is
# adjustable (needed by FreeBSD's "gptloader" on a freebsd-boot UUID
# partition), probably needed by Haiku as well since stage1 is 2 sectors
# big), but beware that int13h 4200h only supports up to 127 (0x7f) sectors,
# or a loader of almost 64KiB; so don't specify a larger number. This is not
# checked by the code, again because of space constraints.
#

				.set LOAD,0x7c00				# Load address
				.set EXEC,0x600					# Execution address
				.set MAGIC,0xaa55				# Magic: bootable
				.set SECSIZE,0x200				# Size of a single disk sector
				.set DISKSIG,440				# Disk signature offset
				.set PT_OFF,0x1be				# Partition table

				.set STACK,EXEC+SECSIZE*4		# Stack address

				.set FL_PACKET,0x80				# Flag: try EDD

				.set GPT_ADDR,STACK				# GPT header address
				.set GPT_SIG,0
				.set GPT_SIG_0,0x20494645
				.set GPT_SIG_1,0x54524150
				.set GPT_MYLBA,24
				.set GPT_PART_LBA,72
				.set GPT_NPART,80
				.set GPT_PART_SIZE,84
				.set PART_ADDR,GPT_ADDR+SECSIZE	# GPT partition array addr.
				.set PART_TYPE,0
				.set PART_START_LBA,32
				.set PART_END_LBA,40

				.set NHRDRV,0x475				# Number of hard drives

				.globl start					# Entry point
				.code16

#
# Setup the segment registers for flat addressing and setup the stack.
#
start:			cld							# String ops inc
				call setup					# Used by the two loader types,
											#  GPT and MBR; must come
											#  after relocation

				movw $STACK,%sp				# Set up GPT stack

#
# Relocate ourself to a lower address so that we have more room to load
# other sectors.
#
reloc:			movw $main_gpt-EXEC+LOAD,%si			# Source
				movw $main_gpt,%di						# Destination
				movw $SECSIZE-(main_gpt-start),%cx		# Byte count
				rep										# Relocate
				movsb									#  code

#
# Jump to the relocated code.
#
				jmp main_gpt-LOAD+EXEC		# To relocated code


main_gpt:		call validate_drv			# validate_drv is also
											#  shared between GPT
											#  and MBR loaders

#
# Load the primary GPT header from LBA 1 and verify signature.
#
load_gpt_hdr:	movw $GPT_ADDR,%bx
				movw $lba,%si
				movw $1,%cx				# Read 1 sector
				call read
				cmpl $GPT_SIG_0,GPT_ADDR+GPT_SIG
				jnz try_mbr				# If invalid GPT header
				cmpl $GPT_SIG_1,GPT_ADDR+GPT_SIG+4
				jnz try_mbr				# Fluke :( Try MBR now

#
# GPT is valid. Load a partition table sector from disk and look for a
# partition matching the UUID found in boot_uuid.
#
load_part:		movw $GPT_ADDR+GPT_PART_LBA,%si
				movw $PART_ADDR,%bx
				movw $1,%cx				# Read 1 sector
				call read
scan:			movw %bx,%si			# Compare partition UUID
				movw $boot_uuid,%di		#  with Haiku boot UUID
				movb $0x10,%cl
				repe cmpsb
				jnz next_part			# Didn't match, next partition
#
# We found a partition. Load it into RAM starting at 0x7c00.
#
				movw %bx,%di			# Save partition pointer in %di
				leaw PART_START_LBA(%di),%si
				movw $LOAD/16,%bx
				movw %bx,%es
				xorw %bx,%bx

load_bootcode:	push %si				# Save %si

										# Load up to 127 sectors into
				movb sectors,%cl		#  physical address 0x7c00
				call read				#  (Notice that %cx is zero due
										#  to "repe cmpsb" above)

start_loader:	xorw %ax,%ax
				movw %ax,%es			# Reset %es to zero
				jmp LOAD				# Jump to boot code

#
# Move to the next partition.  If we walk off the end of the sector, load
# the next sector.
#
next_part:		decl GPT_ADDR+GPT_NPART			# Was this the last partition?
				jz try_mbr						# UUID boot signature not found
				movw GPT_ADDR+GPT_PART_SIZE,%ax
				addw %ax,%bx					# Next partition
				cmpw $PART_ADDR+0x200,%bx		# Still in sector?
				jb scan
				incl GPT_ADDR+GPT_PART_LBA		# Next sector
				adcl $0,GPT_ADDR+GPT_PART_LBA+4
				jmp load_part

#
# If loading a Haiku partition from a GPT fails, try booting a MBR partition.
#
try_mbr:		call setup					# Start afresh
				movw $LOAD,%sp				# Set up MBR stack
main_mbr:		xorw %si,%si				# No active partition
				movw $partbl,%bx			# Partition table
				movb $0x4,%cl				# Number of entries
main_mbr.1:		cmpb %ch,(%bx)				# Null entry?
				je main_mbr.2				# Yes
				jg err_pt					# If 0x1..0x7f
				testw %si,%si				# Active already found?
				jnz err_pt					# Yes
				movw %bx,%si				# Point to active
main_mbr.2:		addb $0x10,%bl				# Till
				loop main_mbr.1				#  done
				testw %si,%si				# Active found?
				jnz main_mbr.3				# Yes
				int $0x18					# BIOS: Diskless boot

#
# Ok, we've found a possible active partition.  Check to see that the drive
# is a valid hard drive number.
#
main_mbr.3:	call validate_drv

#
# Ok, now that we have a valid drive and partition entry, load the CHS from
# the partition entry and read the sector from the disk.
#
main_mbr.4:		movw %sp,%di				# Save stack pointer
				movb 0x1(%si),%dh			# Load head
				movw 0x2(%si),%cx			# Load cylinder:sector
				movw $LOAD,%bx				# Transfer buffer
				testb $FL_PACKET,flags		# Try EDD?
				jz main_mbr.6				# No.
				pushw %cx					# Save %cx (For CHS load)
				pushw %bx					# Save %bx (Ditto)
				movw $0x55aa,%bx			# Magic
				movb $0x41,%ah				# BIOS:	EDD extensions
				int $0x13					#  present?
				jc main_mbr.5				# No.
				cmpw $0xaa55,%bx			# Magic ok?
				jne main_mbr.5				# No.
				testb $0x1,%cl				# Packet mode present?
				jz main_mbr.5				# No.
				popw %bx					# Restore %bx
				pushl $0x0					# Set the LBA
				pushl 0x8(%si)				#  address
				pushw %es					# Set the address of
				pushw %bx					#  the transfer buffer
				pushw $0x1					# Read 1 sector
				pushw $0x10					# Packet length
				movw %sp,%si				# Packet pointer
				movw $0x4200,%ax			# BIOS:	LBA Read from disk
				jmp main_mbr.7				# Skip the CHS setup
main_mbr.5:		popw %bx					# Restore %bx (For CHS load)
				popw %cx					# Restore %cx (For CHS load)
main_mbr.6:		movw $0x201,%ax				# BIOS: Read from disk
main_mbr.7:		int $0x13					# Call the BIOS
				movw %di,%sp				# Restore stack
				jc err_rd					# If error

#
# Now that we've loaded the bootstrap, check for the 0xaa55 signature.  If it
# is present, execute the bootstrap we just loaded.
#
				cmpw $MAGIC,0x1fe(%bx)		# Bootable?
				jne err_noboot				# No
				jmp start_loader			# Invoke bootstrap


################################################################################

#
# Clear registers and set up stack/segments. Called at the start of both GPT and
# MBR loaders.
#
setup:			xorw %cx,%cx
				xorw %ax,%ax				# Zero
				movw %ax,%es				# Address
				movw %ax,%ds				#  data
				movw %ax,%ss
				ret

#
# Validate drive number in %dl.
#
validate_drv:	cmpb $0x80,%dl				# Drive valid?
				jb validate_drv.1			# No
				movb NHRDRV,%dh				# Calculate the highest
				addb $0x80,%dh				#  drive number available
				cmpb %dh,%dl				# Within range?
				jb validate_drv.2			# Yes
validate_drv.1:	movb $0x80,%dl				# Assume drive 0x80
validate_drv.2:	ret

#
# Load a sector (64-bit LBA at %si) from disk %dl into %es:%bx by creating
# a EDD packet on the stack and passing it to the BIOS.  Trashes %ax and %si.
#
read:			pushl 0x4(%si)				# Set the LBA
				pushl 0x0(%si)				#  address
				pushw %es					# Set the address of
				pushw %bx					#  the transfer buffer
				pushw %cx					# Read up to 127 sectors
				pushw $0x10					# Packet length
				movw %sp,%si				# Packet pointer
				movw $0x4200,%ax			# BIOS:	LBA Read from disk
				int $0x13					# Call the BIOS
				add $0x10,%sp				# Restore stack
				jc try_mbr					# In case of error, try the
											#  MBR code which supports
											#  other int13h calls
				ret

#
# Various error message entry points.
#
err_pt:			movw $msg_pt,%si			# "Bad partition
				jmp putstr					#  table"

err_rd:			movw $msg_rd,%si			# "IO Error"
				jmp putstr

err_noboot:		movw $msg_noboot,%si		# "Missing System
				jmp putstr					#  Bootloader"
#
# Output an ASCIZ string to the console via the BIOS.
#
putstr.0:		movw $0x7,%bx				# Page:attribute
				movb $0xe,%ah				# BIOS: Display
				int $0x10					#  character
putstr:			lodsb						# Get character
				testb %al,%al				# End of string?
				jnz putstr.0				# No
putstr.1:		hlt							# Await reset
				jmp putstr.1				#  (but avoid melting the CPU)

msg_pt:			.asciz "Bad Partition Table"
msg_rd:			.asciz "IO Error"
msg_noboot:		.asciz "Missing System Bootloader"

				.org DISKSIG-1-1-16-8,0xff	# -flags-sectors-boot_uuid-lba

flags:			.byte 0x80					# Try EDD packet mode

sectors:		.byte 0x02					# Number of sectors to load
											#  from the target partition

boot_uuid:		.long 0x42465331			# 'BFS1' (formally, UUID time-low)
				.word 0x3ba3				# UUID time-mid
				.word 0x10f1				# UUID time-high & version (v1)
				.byte 0x80					# UUID DCE 1.1 variant
				.byte 0x2a					# '*' (formally, UUID clock-seq-low)
				.byte 0x48					# 'H'
				.byte 0x61					# 'a'
				.byte 0x69					# 'i'
				.byte 0x6b					# 'k'
				.byte 0x75					# 'u'
				.byte 0x21					# '!'

lba:			.quad 1						# LBA of GPT header

				.org DISKSIG,0x90			# Won't pad (code is packed tight!)
											#  but here for documentation

sig:			.long 0						# OS Disk Signature
				.word 0						# "Unknown" in PMBR

partbl:			.fill 0x10,0x4,0x0			# Partition table
				.word MAGIC					# Magic number
